from argparse import ArgumentParser, Namespace
import importlib
from pathlib import Path

from migratik.config import get_config, Config
from migratik.migrator import Migrator
from migratik.backends.backend import AbstractBackend
from migratik.errors import MigratikError, ConfigError


_CONFIG_TEMPLATE = (
    "migrations:"
    '\n    path: "..."'
    "\n\nbackend:"
    '\n    class: "..."'
    "\n    kwargs:"
    "\n        ..."
    "\n"
)


def process_command() -> None:
    parser = ArgumentParser(prog="migratik")
    subparsers = parser.add_subparsers()

    init_parser = subparsers.add_parser("init")
    init_parser.set_defaults(func=process_init)

    status_parser = subparsers.add_parser("status")
    status_parser.set_defaults(func=process_status)

    create_parser = subparsers.add_parser("create")
    create_parser.set_defaults(func=process_create)

    upgrade_parser = subparsers.add_parser("upgrade")
    upgrade_parser.add_argument("-v", "--version", type=int)
    upgrade_parser.set_defaults(func=process_upgrade)

    downgrade_parser = subparsers.add_parser("downgrade")
    downgrade_parser.add_argument("-v", "--version", type=int, required=True)
    downgrade_parser.set_defaults(func=process_downgrade)

    namespace = parser.parse_args()

    try:
        namespace.func(namespace)
    except AttributeError:
        print("You didn't specify a command.")


def process_init(namespace: Namespace) -> None:
    config_path = Path.cwd() / "migratik.yml"

    if config_path.exists():
        print("Migratik has already been initialized.")

        return

    config_path.write_text(_CONFIG_TEMPLATE, encoding="UTF-8")


def process_status(namespace: Namespace) -> None:
    try:
        config = _get_config()
        migrator = Migrator(config.migrations.path)
        versions = migrator.get_versions()
        backend = _get_backend(config)

        if versions:
            latest_version = versions[-1]
        else:
            latest_version = "—"

        current_version = backend.get_version() or "—"
    except MigratikError as error:
        print(error)
    else:
        print(
            f"Current version: {current_version}"
            f"\nLatest version: {latest_version}"
        )


def process_create(namespace: Namespace) -> None:
    try:
        config = _get_config()
        migrator = Migrator(config.migrations.path)
        path = migrator.create_migration()
    except MigratikError as error:
        print(error)
    else:
        print(f"Migration {str(path)!r} created.")


def process_upgrade(namespace: Namespace) -> None:
    try:
        config = _get_config()
        migrator = Migrator(config.migrations.path)
        backend = _get_backend(config)
        migrator.upgrade(backend, namespace.version)
    except MigratikError as error:
        print(error)


def process_downgrade(namespace: Namespace) -> None:
    try:
        config = _get_config()
        migrator = Migrator(config.migrations.path)
        backend = _get_backend(config)
        migrator.downgrade(backend, namespace.version)
    except MigratikError as error:
        print(error)


def _get_config() -> Config:
    try:
        return get_config(
            Path.cwd() / "migratik.yml"
        )
    except FileNotFoundError:
        raise MigratikError(
            "Configuration file was not found. "
            "Initialize using «migratik init» command."
        ) from None


def _get_backend(config: Config) -> AbstractBackend:
    if config.backend is None:
        raise ConfigError("There is no backend data in configuration!")

    class_ = config.backend.class_

    if "." in class_:
        module, class_ = class_.rsplit(".", 1)
    else:
        module = "migratik"

    module = importlib.import_module(module)
    class_ = getattr(module, class_)

    return class_(**config.backend.kwargs)



from abc import ABC, abstractmethod
from typing import Optional, Any


class AbstractCursor(ABC):

    @abstractmethod
    def execute_query(self, query: str, parameters: Optional[dict[str, Any]] = None) -> None:
        pass

    @abstractmethod
    def fetch_row(self) -> Optional[dict[str, Any]]:
        pass

    @abstractmethod
    def close(self) -> None:
        pass


class AbstractCursorContext(ABC):

    def __init__(self):
        self._cursor: Optional[AbstractCursor] = None

    def __enter__(self) -> AbstractCursor:
        self._cursor = self.get_cursor()

        return self._cursor

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._cursor.close()

    @abstractmethod
    def get_cursor(self) -> AbstractCursor:
        pass



from typing import Optional


class VersionList:

    def __init__(self, versions: list[int]):
        self._versions = sorted(versions)
        self._indexes = {version: index for index, version in enumerate(self._versions)}

    def __iter__(self):
        return iter(self._versions)

    def get_for_upgrade(self, current_version: int, version: Optional[int] = None) -> list[int]:
        current_version_index = self._indexes[current_version]

        if version is not None:
            if current_version > version:
                raise ValueError(
                    f"Current version {current_version} is greater "
                    f"than upgrade version {version}!"
                )

            version_index = self._indexes[version]

            return self._versions[current_version_index + 1:version_index + 1]

        return self._versions[current_version_index + 1:]

    def get_for_downgrade(self, current_version: int, version: int) -> list[int]:
        if current_version < version:
            raise ValueError(
                f"Current version {current_version} is smaller "
                f"than downgrade version {version}!"
            )

        current_version_index = self._indexes[current_version]
        version_index = self._indexes[version]

        return self._versions[current_version_index:version_index:-1]



from dataclasses import dataclass
from typing import Any, Union, Optional
from pathlib import Path

import yaml
from yaml import Loader


@dataclass
class MigrationSection:
    path: Path


@dataclass
class BackendSection:
    class_: str
    kwargs: dict[str, Any]


@dataclass
class Config:
    migrations: MigrationSection
    backend: Optional[BackendSection] = None


def get_config(path: Union[str, Path]) -> Config:
    path = Path(path)

    with path.open(encoding="UTF-8") as file:
        data = yaml.load(file, Loader)

    backend_section = data.get("backend")

    if backend_section is not None:
        backend_section = BackendSection(
            class_=backend_section["class"],
            kwargs=backend_section.get("kwargs", {})
        )

    return Config(
        migrations=MigrationSection(
            path=Path(data["migrations"]["path"])
        ),
        backend=backend_section
    )



from pathlib import Path
from typing import Union, Optional
from datetime import datetime
import enum
from enum import IntEnum

from migratik.types import Migration
from migratik.backends.backend import AbstractBackend
from migratik.version_list import VersionList
from migratik.errors import (
    MigrationError,
    MigrationPathError,
    MigrationParsingError
)


_UPGRADE_COMMENT = "-- Upgrade:"
_DOWNGRADE_COMMENT = "-- Downgrade:"
_DIGITS = frozenset("0123456789")


class ParsingState(IntEnum):
    INITIAL = enum.auto()
    UPGRADE_COLLECTING = enum.auto()
    DOWNGRADE_COLLECTING = enum.auto()


class Migrator:

    def __init__(self, path: Union[str, Path]):
        self.path = Path(path)

    def create_migration(self) -> Path:
        if not self._check_path():
            self.path.mkdir()

        current_datetime = datetime.utcnow()
        version = int(
            current_datetime.strftime("%Y%m%d%H%M%S%f")
        )
        formatted_current_datetime = current_datetime.strftime('%Y-%m-%d %H:%M:%S')
        content = (
            f"-- Version {version}."
            f"\n-- Created at {formatted_current_datetime} UTC."
            f"\n\n\n{_UPGRADE_COMMENT}\n"
        )

        if any(i.is_file() and _check_name(i.name) for i in self.path.iterdir()):
            content += f"\n\n\n{_DOWNGRADE_COMMENT}\n"

        content += "\n"
        path = self._get_migration_path(version)
        path.write_text(content, encoding="UTF-8")

        return path

    def get_versions(self) -> list[int]:
        if not self._check_path():
            raise MigrationPathError(
                f"Migration directory {str(self.path)!r} does not exist!"
            )

        return sorted(
            int(i.stem)
            for i in self.path.iterdir()
            if i.is_file() and _check_name(i.name)
        )

    def get_migration(self, version: int) -> Migration:
        if version < 0:
            raise ValueError("Version cannot be negative!")

        if not self._check_path():
            raise MigrationPathError(
                f"Migration directory {str(self.path)!r} does not exist!"
            )

        path = self._get_migration_path(version)

        if not path.exists():
            raise MigrationPathError(
                f"Migration file {str(path)!r} does not exist!"
            )

        return self._get_migration(version)

    def upgrade(self, backend: AbstractBackend, version: Optional[int] = None) -> None:
        if (version is not None) and (version < 0):
            raise ValueError("Version cannot be negative!")

        versions = self.get_versions()

        if not versions:
            raise MigrationError("No migration files!")

        versions = VersionList(versions)

        with backend.connect() as connection:
            with connection.get_cursor() as cursor:
                if backend.check_migration_table(cursor):
                    last_migration = backend.get_last_migration(cursor)

                    if last_migration is not None:
                        current_version = last_migration["version"]

                        if version is not None:
                            if current_version == version:
                                return

                        versions = versions.get_for_upgrade(current_version, version)
                else:
                    backend.create_migration_table(cursor)

                migrations = [self._get_migration(i) for i in versions]
                processed_migrations = []

                for migration in migrations:
                    try:
                        cursor.execute_query(migration.upgrade_queries)
                    except Exception:  # noqa
                        connection.rollback_transaction()
                        connection.start_transaction()

                        for processed_migration in reversed(processed_migrations):
                            cursor.execute_query(processed_migration.downgrade_queries)
                            backend.delete_migration(cursor, version=processed_migration.version)
                            connection.commit_transaction()
                            connection.start_transaction()

                        raise
                    else:
                        backend.add_migration(cursor, version=migration.version)
                        connection.commit_transaction()
                        processed_migrations.append(migration)
                        connection.start_transaction()

    def downgrade(self, backend: AbstractBackend, version: int) -> None:
        if version < 0:
            raise ValueError("Version cannot be negative!")

        versions = self.get_versions()

        if not versions:
            raise MigrationError("No migration files!")

        versions = VersionList(versions)

        with backend.connect() as connection:
            with connection.get_cursor() as cursor:
                if not backend.check_migration_table(cursor):
                    raise MigrationError("Migration table has not been initialized!")

                last_migration = backend.get_last_migration(cursor)

                if last_migration is None:
                    raise MigrationError("There is not single migration in migration table!")

                current_version = last_migration["version"]

                if current_version == version:
                    return

                versions = versions.get_for_downgrade(current_version, version)
                migrations = [self._get_migration(i) for i in versions]
                processed_migrations = []

                for migration in migrations:
                    try:
                        cursor.execute_query(migration.downgrade_queries)
                    except Exception:  # noqa
                        connection.rollback_transaction()
                        connection.start_transaction()

                        for processed_migration in reversed(processed_migrations):
                            cursor.execute_query(processed_migration.upgrade_queries)
                            backend.add_migration(cursor, version=processed_migration.version)
                            connection.commit_transaction()
                            connection.start_transaction()

                        raise
                    else:
                        backend.delete_migration(cursor, version=migration.version)
                        connection.commit_transaction()
                        processed_migrations.append(migration)
                        connection.start_transaction()

    def _check_path(self) -> bool:
        if self.path.exists():
            if not self.path.is_dir():
                raise MigrationPathError(f"Path {str(self.path)!r} is not directory!")

            return True

        return False

    def _get_migration(self, version: int) -> Migration:
        path = self._get_migration_path(version)
        upgrade_lines = []
        downgrade_lines = []
        state = ParsingState.INITIAL

        with path.open(encoding="UTF-8") as file:
            for line in file:
                striped_line = line.strip()

                if state is ParsingState.INITIAL:
                    if striped_line == _UPGRADE_COMMENT:
                        state = ParsingState.UPGRADE_COLLECTING
                    elif striped_line == _DOWNGRADE_COMMENT:
                        raise MigrationParsingError(
                            f"Upgrade block missed ({path})!"
                        )
                elif state is ParsingState.UPGRADE_COLLECTING:
                    if striped_line == _DOWNGRADE_COMMENT:
                        state = ParsingState.DOWNGRADE_COLLECTING
                    elif striped_line == _UPGRADE_COMMENT:
                        raise MigrationParsingError(
                            f"Extra upgrade block was found ({path})!"
                        )
                    else:
                        upgrade_lines.append(
                            line.rstrip()
                        )
                elif state is ParsingState.DOWNGRADE_COLLECTING:
                    if striped_line == _UPGRADE_COMMENT:
                        raise MigrationParsingError(
                            f"Extra upgrade block was found ({path})!"
                        )
                    elif striped_line == _DOWNGRADE_COMMENT:
                        raise MigrationParsingError(
                            f"Extra downgrade block was found ({path})!"
                        )
                    else:
                        downgrade_lines.append(
                            line.rstrip()
                        )

        return Migration(
            version=version,
            upgrade_queries="\n".join(upgrade_lines).rstrip(),
            downgrade_queries="\n".join(downgrade_lines).rstrip()
        )

    def _get_migration_path(self, version: int) -> Path:
        return self.path / f"{version}.sql"


def _check_name(file_name: str) -> bool:
    try:
        name, extension = file_name.rsplit(".", 1)
    except ValueError:
        return False

    return (
        (extension == "sql")
        and bool(name)
        and all(i in _DIGITS for i in name)
    )



from .backends import (
    AbstractBackend,
    AbstractConnection,
    AbstractConnectionContext,
    AbstractCursor,
    AbstractCursorContext,
    Psycopg2Backend
)
from .migrator import Migrator


__all__ = [
    "AbstractBackend",
    "AbstractConnection",
    "AbstractConnectionContext",
    "AbstractCursor",
    "AbstractCursorContext",
    "Psycopg2Backend",
    "Migrator"
]



from .backend import AbstractBackend
from .connection import AbstractConnection, AbstractConnectionContext
from .cursor import AbstractCursor, AbstractCursorContext
from .psycopg2.backend import Psycopg2Backend


__all__ = [
    "AbstractBackend",
    "AbstractConnection",
    "AbstractConnectionContext",
    "AbstractCursor",
    "AbstractCursorContext",
    "Psycopg2Backend"
]



from abc import ABC, abstractmethod
from typing import Optional

from migratik.backends.cursor import AbstractCursorContext


class AbstractConnection(ABC):

    @abstractmethod
    def start_transaction(self) -> None:
        pass

    @abstractmethod
    def get_cursor(self) -> AbstractCursorContext:
        pass

    @abstractmethod
    def commit_transaction(self) -> None:
        pass

    @abstractmethod
    def rollback_transaction(self) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass


class AbstractConnectionContext(ABC):

    def __init__(self):
        self._connection: Optional[AbstractConnection] = None

    def __enter__(self) -> AbstractConnection:
        self._connection = self.get_connection()
        self._connection.start_transaction()

        return self._connection

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self._connection.rollback_transaction()
        else:
            self._connection.commit_transaction()

        self._connection.close()

    @abstractmethod
    def get_connection(self) -> AbstractConnection:
        pass





class MigratikError(Exception):
    pass


class MigrationError(MigratikError):
    pass


class MigrationPathError(MigratikError):
    pass


class MigrationParsingError(MigratikError):
    pass


class MigrationTableError(MigratikError):
    pass


class ConfigError(MigratikError):
    pass



from typing import Any, Optional

from migratik.backends.cursor import AbstractCursor, AbstractCursorContext


class Psycopg2Cursor(AbstractCursor):

    def __init__(self, cursor):
        self._cursor = cursor

    def execute_query(self, query: str, parameters: Optional[dict[str, Any]] = None) -> None:
        self._cursor.execute(query, parameters)

    def fetch_row(self) -> Optional[dict[str, Any]]:
        row = self._cursor.fetchone()

        if row is not None:
            return dict(row)

    def close(self) -> None:
        self._cursor.close()


class Psycopg2CursorContext(AbstractCursorContext):

    def __init__(self, connection):
        super().__init__()
        self.connection = connection

    def get_cursor(self) -> Psycopg2Cursor:
        return Psycopg2Cursor(
            self.connection.cursor()
        )



from abc import ABC, abstractmethod
from typing import Optional, Any

from migratik.backends.connection import AbstractConnectionContext
from migratik.backends.cursor import AbstractCursor
from migratik.errors import MigrationTableError


class AbstractBackend(ABC):

    def __init__(self, migration_table: Optional[str] = None):
        self.migration_table = migration_table or "migratik"
        self.check_package()

    @abstractmethod
    def check_package(self) -> None:
        pass

    @abstractmethod
    def connect(self) -> AbstractConnectionContext:
        pass

    @abstractmethod
    def check_migration_table(self, cursor: AbstractCursor) -> bool:
        pass

    @abstractmethod
    def create_migration_table(self, cursor: AbstractCursor) -> None:
        pass

    @abstractmethod
    def add_migration(self, cursor: AbstractCursor, *, version: int) -> None:
        pass

    @abstractmethod
    def delete_migration(self, cursor: AbstractCursor, *, version: int) -> None:
        pass

    @abstractmethod
    def get_last_migration(self, cursor: AbstractCursor) -> Optional[dict[str, Any]]:
        pass

    def initialize_migration_table(self, version: int) -> None:
        with self.connect() as connection:
            with connection.get_cursor() as cursor:
                if self.check_migration_table(cursor):
                    raise MigrationTableError(
                        f"Migration table {self.migration_table!r} already exists!"
                    )

                self.create_migration_table(cursor)
                self.add_migration(cursor, version=version)

    def get_version(self) -> Optional[int]:
        with self.connect() as connection:
            with connection.get_cursor() as cursor:
                if self.check_migration_table(cursor):
                    last_migration = self.get_last_migration(cursor)

                    if last_migration is not None:
                        return last_migration["version"]



from typing import Optional, Any

from migratik.backends.backend import AbstractBackend
from migratik.backends.psycopg2.connection import Psycopg2ConnectionContext
from migratik.backends.psycopg2.cursor import Psycopg2Cursor


class Psycopg2Backend(AbstractBackend):

    def __init__(
        self,
        database: str,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        migration_table: Optional[str] = None
    ):
        super().__init__(migration_table=migration_table)
        self.database = database
        self.user = user
        self.password = password
        self.host = host
        self.port = port

    def check_package(self) -> None:
        try:
            import psycopg2
        except ImportError:
            raise ImportError(
                "To use Psycopg2Backend you need to install «psycopg2»:"
                "\nhttps://www.psycopg.org/docs/install.html"
            ) from None

    def connect(self) -> Psycopg2ConnectionContext:
        return Psycopg2ConnectionContext(
            database=self.database,
            user=self.user,
            password=self.password,
            host=self.host,
            port=self.port
        )

    def check_migration_table(self, cursor: Psycopg2Cursor) -> bool:
        cursor.execute_query(
            f"""
            SELECT
                EXISTS (
                    SELECT
                        1
                    FROM
                        "information_schema"."tables"
                    WHERE
                        "table_name" = '{self.migration_table}'
                )
                AS "result";
            """
        )

        return cursor.fetch_row()["result"]

    def create_migration_table(self, cursor: Psycopg2Cursor) -> None:
        cursor.execute_query(
            f"""
            CREATE TABLE "{self.migration_table}" (
                "version"                   NUMERIC
                                            PRIMARY KEY,
                "creation_timestamp"        TIMESTAMP WITH TIME ZONE
                                            NOT NULL
                                            DEFAULT (NOW())
            );
            """
        )

    def add_migration(self, cursor: Psycopg2Cursor, *, version: int) -> None:
        cursor.execute_query(
            f"""
            INSERT INTO
                "{self.migration_table}" ("version")
            VALUES
                (%(version)s);
            """,
            {
                "version": version
            }
        )

    def delete_migration(self, cursor: Psycopg2Cursor, *, version: int) -> None:
        cursor.execute_query(
            f"""
            DELETE FROM
                "{self.migration_table}"
            WHERE
                "version" = %(version)s;
            """,
            {
                "version": version
            }
        )

    def get_last_migration(self, cursor: Psycopg2Cursor) -> Optional[dict[str, Any]]:
        cursor.execute_query(
            f"""
            SELECT
                "version", "creation_timestamp"
            FROM
                "{self.migration_table}"
            ORDER BY
                "version" DESC
            LIMIT
                1;
            """
        )

        return cursor.fetch_row()



from typing import Optional

from migratik.backends.connection import AbstractConnection, AbstractConnectionContext
from migratik.backends.psycopg2.cursor import Psycopg2CursorContext


class Psycopg2Connection(AbstractConnection):

    def __init__(self, connection):
        self._connection = connection

    def start_transaction(self) -> None:
        pass

    def get_cursor(self) -> Psycopg2CursorContext:
        return Psycopg2CursorContext(self._connection)

    def commit_transaction(self) -> None:
        self._connection.commit()

    def rollback_transaction(self) -> None:
        self._connection.rollback()

    def close(self) -> None:
        self._connection.close()


class Psycopg2ConnectionContext(AbstractConnectionContext):

    def __init__(
        self,
        database: str,
        user: Optional[str] = None,
        password: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None
    ):
        super().__init__()
        self.database = database
        self.user = user
        self.password = password
        self.host = host
        self.port = port

    def get_connection(self) -> Psycopg2Connection:
        import psycopg2
        from psycopg2.extras import RealDictCursor

        return Psycopg2Connection(
            psycopg2.connect(
                database=self.database,
                user=self.user,
                password=self.password,
                host=self.host,
                port=self.port,
                cursor_factory=RealDictCursor
            )
        )
